//-----------------------------------------------------------------------------
// GA2K
//
// ISDL, Doshisha University Distributed Genetic Algorithm Software
// Version 1.1
//-----------------------------------------------------------------------------
// Parallel Genetic Algorithm Team of Intelligent Systems Design Laborato
// ry, Doshisha University.
//
// Version 0.1: 1999/05/05
// Copyright (C) 1999 Jun-ichi Yoshida and Ikki Ohmukai, All rights reserved.
// Version 1.0: 2001/12/13
// Copyright (C) 2001 Jun-ichi Yoshida, Tomoyuki Hiroyasu and Mitsunori Miki,
// All rights reserved.
// Version 1.1: 2002/01/17
// Copyright (C) 2002 Jun-ichi Yoshida, Masaki Sano, Tomoyuki Hiroyasu and Mitsunori Miki, All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
//    notce and this list of conditions.
// 2. Redistributions in binary form must reproduce the above  copyright
//    notice and this list of conditions in the documentation and/or other
//    materials provided with the distribution.
// 3. The names of copyrighters of this software  should not be used to
//    endorse or promote products derived from this software without specific
//    prior written permission.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// ga2k_dmachine.cpp
//-----------------------------------------------------------------------------

#include <cstdio>
#include <cstdlib>
#include <cfloat>
#include <vector>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <cmath>

using namespace std;

extern "C" {
#include <getopt.h>
#include <unistd.h>
#include <dmachinemon/dmachinemon.h>
}
#include "ga2k_dmachine.h"
#include "randring.h"


dm_commandoption cdat;

//=============================================================================
//functions
//=============================================================================

static const char * mytag="ga2kdm";

/**
   @Returns 0 on success
 */
static int send_gene(const Gene& gene)
{
  char *sendbuf = (char*)malloc (gene.bitString.size() + 1);
  memcpy (sendbuf, (void*)(gene.bitString.begin()), gene.bitString.size());
  sendbuf[gene.bitString.size()] = 0;
  for (unsigned int i = 0; i < gene.bitString.size(); ++i )
    sendbuf[i]+='0';

  DNAS_sendinfo(mytag, sendbuf, &cdat);
  //cerr << sendbuf << endl;	// debug
  free (sendbuf);
  return 0;
}


/**
   @Returns 0 on success
 */
static int recv_gene(Gene& gene)
{
  char ** m;  
  char *buf = (char*)(gene.bitString.begin());
  int count = gene.bitString.size();
  int retval = 0;

  m = DNAS_gatherinfo("ga2kdm", &cdat);
  if (m)
    {
      int i, num_candidates=0;
      for (i=0; m[i]; i++) 
	{
	  num_candidates = i;
	}
      if (num_candidates >= 1)
	{
	  memcpy(buf, m[genrand_int(num_candidates)], count);
	  for (int i = 0; i < count; ++i )
	    buf[i]-= '0';
	}
      else
	{
	  retval = 1;
	}
      DNAS_freeinfo(m);
    }
  else
    retval = 1;			// failure


  return retval;
}


//=============================================================================
//class DMGlobe
//=============================================================================

DMGlobe::DMGlobe(int *pargc, char*** pargv) 
{
  DNAS_Init(pargc, pargv, &cdat);
}

DMGlobe::~DMGlobe()
{
  DNAS_Finalize(&cdat);
}

void DMGlobe::initialize()
{
  // Default settings

  fTotalPopSize = false;
  totalPopulationSize = 400;

  island_0.islandId		= 0; // FIXMEdummy treatment

  // parameters for chromosome
  island_0.geneLength		= 100;
  island_0.variableLength	= 10;
  island_0.numVariable		= 10;

  fInitialPopulation			= 0; // 0: random,  1: from file(design variables)

  // parameters for objective function
  island_0.functionType		= 0;
  island_0.openMax			= 5.12;
  island_0.closedMin			= -5.12;
  island_0.delta				= 0.01;
  island_0.finValue			= -0.00000001;

  // canonical GA parameters
  island_0.populationSize		= 10; // in each subpopulations
  island_0.crossoverRate		= 1.0;
  island_0.mutationRate		= 0.01;
  island_0.numElite			= 5;
  island_0.tournamentSize    = 4;
  island_0.selectionType     = 0;

  // DGA parameters
  island_0.migrationRate		= 0.5;
  migrationInterval			= 5;


  island_0.numCrossoverPoint	= 2;
  island_0.typeOfEncode		= 0;


  island_0.numEvaluate		= 0;
  maxGeneration		= 1000;


  maxTrial			= 1;
  sumMaxFitness		= 0;
  sumGeneration		= 0;
  numConvergent		= 0;
  sumTotalEvaluate	= 0;
  fStop				= true;
  fOutput				= false;
}

void DMGlobe::initialize(int argc, char *argv[])
{
  initialize();
  int opt;

  // command line
  while (-1!=(opt=getopt(argc, argv, "b:i:c:f:I:m:M:g:op:P:z:S:s:")))
    {
      const char *p = optarg;
      switch(opt)
	{

	case 'b':
	  island_0.geneLength = atoi(p);
	  island_0.numVariable = island_0.geneLength / island_0.variableLength;
	  break;

	case 'i':
	  numIsland = atoi(p);
	  break;

	case 'c':
	  island_0.crossoverRate = atof(p);
	  break;

	case 'f':
	  island_0.functionType = atoi(p);
	  if(atoi(p) == 0){
	    island_0.closedMin = -5.12;
	    island_0.openMax = 5.12;
	    island_0.delta = 0.01;
	  }
	  else if(atoi(p) == 1 || atoi(p) == 2){
	    island_0.closedMin = -512;
	    island_0.openMax = 512;
	    island_0.delta = 1;
	  }
	  else if(atoi(p) == 3){
	    island_0.closedMin = -2.048;
	    island_0.openMax = 2.048;
	    island_0.delta = 0.004;
	  }
	  else if(atoi(p) == 4){
	    island_0.closedMin = -64;
	    island_0.openMax = 64;
	    island_0.delta = 0.125;
	  }
	  break;

	case 'I':
	  migrationInterval = atoi(p);
	  break;

	case 'm':
	  island_0.mutationRate = atof(p);
	  break;

	case 'M':
	  island_0.migrationRate = atof(p);
	  break;

	case 'g':
	  maxGeneration = atoi(p);
	  break;

	case 'o':
	  fOutput = true;
	  break;

	case 'p':
	  island_0.populationSize = atoi(p);
	  island_0.numElite = island_0.populationSize / 2;
	  break;

	case 'P':
	  totalPopulationSize = atoi(p);
	  fTotalPopSize = true;
	  break;

	case 'z':
	  maxTrial = atoi(p);
	  break;

	case 'S':
	  // 0: random,  1: from file(design variables)
	  fInitialPopulation	= atoi(p);
	  break;

	case 's':
	  // 0: tournament,  1: roulette
	  island_0.selectionType = atoi(p);
	  break;
	default:
	  cerr << "Unknown parameter given" << endl;
	  exit (1);
	  break;
	}
    }
}

void DMGlobe::establish()
{
  if(fTotalPopSize)
    {
      island_0.populationSize = totalPopulationSize / numIsland;
      island_0.numElite = island_0.populationSize / 2;
    }
  island_0.migrantSize	= static_cast<int>(island_0.populationSize * island_0.migrationRate);
  island_0.nativeSize	= island_0.populationSize - island_0.migrantSize;

  populationSize	= island_0.populationSize;
  nativeSize		= island_0.nativeSize;
  migrantSize		= island_0.migrantSize;

  maxFitness		= -DBL_MAX;
  convGeneration	= maxGeneration;
  fConv			= false;
  totalEvaluate	= 0;

  ifstream fin;
  if(fInitialPopulation == 1){
    // ݌vϐl̃t@CI[v
    fin.open(INITIAL_POPULATION_FILE);
    if(!fin.is_open()){
      cerr << __FILE__ << ":(" << __LINE__ << ") can't open file "
	   << INITIAL_POPULATION_FILE << endl;
      exit(1);
    }
  }

  island = island_0;
  island.initialize(0, 0, fin);

  fin.close();

  synchronize();
}

void DMGlobe::emigrate()
{
  // ڏZ̂̓_
  int retval = 0;
  random_shuffle(island.gene.begin(), island.gene.end(), genrand_int);

  for(int j = 0; j < migrantSize; j++){
    send_gene(island.gene[nativeSize + j]);
    retval += recv_gene(island.migrant[j]);
  }
  if (retval )
    {
      cerr << "warning: no info received." << endl;
      usleep (1000000);
    }
}

void DMGlobe::execute()
{
  for(generation = 0; generation < maxGeneration; generation++)
    {
      //if(generation % 10 == 0)
      //cerr << "." << flush;

      if(generation % migrationInterval == 0){
	emigrate();
	island.immigrate();
      }

      island.crossover();
      island.mutate();
      island.evaluate();
      island.elitePreservation();
      
      setMaxFitness();

      island.reproduce();

      if(!fConv && maxFitness > island_0.finValue)
	finalize(true);

      sumFitness[generation] += maxFitness;

      cout << maxFitness <<  " " ;
      for (unsigned int hoge = 0; hoge < island.gene[0].bitString.size(); ++hoge)
	cout << (char)((int)island.gene[1].bitString[hoge] + (int)'0') ;
      cout << endl;
    }
  finalize(false);
}


void DMGlobe::finalize(bool b)
{
  // dmachinemonfł̓t@CiCY͍sȂ(͂ƂȂCƂȂ)
}

//sOɃp[^cerrŏo͂
void DMGlobe::showParameters()
{
  cerr << "\nPopulation     : " << island_0.populationSize * numIsland << endl;
  cerr << "Num island     : " << numIsland << endl;
  cerr << "Pop/island     : " << island_0.populationSize << endl;
  cerr << "init population: " << ((fInitialPopulation == 0) ? "random" : "from file") << endl;
  cerr << "Elite/island   : " << island_0.numElite << endl;
  cerr << "MigRate        : " << island_0.migrationRate << endl;
  cerr << "MigInterval    : " << migrationInterval << endl ;
  cerr << "Pc             : " << island_0.crossoverRate << endl;
  cerr << "Pm             : " << island_0.mutationRate << endl;;
  cerr << "MaxGeneration  : " << maxGeneration << endl;
  
  //֐Ɋւ
  cerr << "Function       : " <<  "f" << island_0.functionType << endl ;
  cerr << "\n";
  cerr << flush;
  synchronize();
}

void DMGlobe::synchronize() const
{
}

void DMGlobe::setMaxFitness()
{
  if (island.bestFitness > maxFitness)
    maxFitness = island.bestFitness;
}

